#!/usr/bin/env python
import spss, spssaux, random, os
"""
compare2files(file1, file2, key=None)
Program to find differences between 2 sav files:
    file1 = filespecs of first file eg file1 = r'c:/temp/file1.sav'
    file2 = filespecs of second file
    key   = None or a list of one or more variables that uniquely identify each case
            eg key = 'eerID eeeID'.split() or
               key = 'countryID cityID householdID personID'.split()
            When key == None, key is assumed to equal (up to) 50 'compatible'
            variables present in both files where 'compatible' means vars are
            both numeric or they are strings of the same length.

Program identifies:
    Variable names present in both files or missing from one file.

    The sum of numeric variables present in both files.
    For vars in both files: Dictionary differences regarding any of varType,
    varFormat, varLabel, valueLabels, missingValues and varLevel are displayed
    in a Pivot Table.

In addition:
    Cases with key without matched case in the other file, or cases for key that
    are duplicates in a file are stored in file "duplicate or missing keys.sav".

    For remaining cases, program matches cases from each file having the same key
    then compares content of every variables. Cases with differences are
    saved in file "records with differences.sav"

Example of use:
# You should first save a copy of compare2files.py to your
# "C:\Python24\Lib\site-packages\" folder

BEGIN PROGRAM.
import compare2files as c2f
c2f.compare2files(file1 = "c:/Program files/SPSS/employee data.sav",
               file2 = "C:/temp/eee data2.sav",
               key = ['id'])
END PROGRAM.

# Potential future improvments :
1) Allow keys containing more than 50 vars

"""

__author__  =  'Raynald Levesque raynald@spsstools.net'
__version__ =  '1.1'


def _getFullDict(fname):
    """ Get complete dictionary info for file fname
    Returns 1) a set containing variable names and 2) dictionary
    """
    spss.Submit(r"GET FILE='%s'." % fname)
    nbCases = spss.GetCaseCount()
    varDict = spssaux.VariableDict()
    varSet = set([id for id in varDict.variables])
    fullDict = {}

    for id in varDict.variables:
        fullDict[id] = (varDict.VariableType(id), varDict.VariableFormat(id)
                    , varDict.VariableLabel(id), varDict.ValueLabels(id)
                    ,varDict.MissingValues(id), varDict.VariableLevel(id) )
    varSet = set([id for id in varDict.variables])
    return nbCases, varSet, fullDict


def _printResults( procTitle, textTitle, myText, myCells, desc ):
    """Print resuls of comparison in a Text Box and Pivot Table"""
    spss.StartProcedure(procTitle)
    spss.TextBlock(textTitle, ''.join(myText))
    if myCells:
        table = spss.BasePivotTable("Dictionary Differences Between File1 and File2",
                                    "OMS table subtype",
                                    caption= ' '.join(['Following characteristics are compared: ',desc]))
        collabs = ['Var name','Dictionary Item','File1','File2']
        table.SimplePivotTable(rowdim = "Row",
                               rowlabels = [n+1 for n in range(len(myCells)/len(collabs))],
                               collabels = collabs,
                               cells = myCells)
    spss.EndProcedure()


def _compareTotals(file1, file2, varNames, nbCases1, nbCases2):
    "Compare totals of varNames in the 2 files. List variables with differences"

    cmd = r"""
        SET MPRINT=YES /PRINTBACK=YES.
        DATASET CLOSE ALL.
        GET FILE="%(file1)s" /KEEP=%(varNames)s.
        COMPUTE fileID = 1.
        DATASET NAME file1.

        GET FILE="%(file2)s" /KEEP=%(varNames)s.
        COMPUTE fileID = 2.
        DATASET NAME file2.

        ADD FILES FILE=file1
                /FILE=file2 .
        AGGREGATE OUTFILE=*
                /BREAK=fileID
                /nb123321=N
                /%(varNames)s  = SUM(%(varNames)s).
        FLIP.

        RENAME VAR (case_lbl var001 var002 = varName file1 file2).
        VAL LAB varName 'nb123321'  'Nb cases'.
        SELECT IF varName <> "fileID".
        COMPUTE delta = SUM( file1, -file2 ).
        COMPUTE deltaAbs = ABS(delta) .
        SORT CASES BY delta(D).
        SELECT IF delta <> 0 .
        FORMATS delta(F8.3) file1 file2(COMMA8).
        EXECUTE /* This EXE is required in order to subsequently read the correct number of cases in */.
        """ % locals()
    print '******************\ncmd=',cmd,'\n******************\n'
    spss.Submit(cmd)
    nbCases = spss.GetCaseCount()
    if nbCases == 0:
        myText = ['The sum of numeric variables present in both files are IDENTICAL.']
    else:
        cmd = r"""
        PRESERVE.
        SET TNUMBERS=LABEL.
        SUMMARIZE
          /TABLES=varName file1 file2 delta
          /FORMAT=VALIDLIST NOCASENUM TOTAL
          /TITLE='Compare 2 files (sum of numeric variables)'
          /MISSING=VARIABLE
          /CELLS=NONE .
        RESTORE. """
        spss.Submit(cmd)
        nbVarsWithDelta = nbCases
        if nbCases1 != nbCases2:
            nbVarsWithDelta = nbCases - 1
        myText = ['%u numeric variables do not have the same totals in both files.' % nbVarsWithDelta,
                  'Refer to the above SUMMARIZE table for variable names and difference amounts.']
    return '\n'.join(myText)

def _compareCases(file1, file2, key, keySource, varNumSet11, varStrSet11):
    """This compares matched cases from the 2 files based on their keys.
    Cases with duplicate keys in a given file, or unmatched key in the other file, are saved to file
    'duplicate or missing keys.sav' (in same folder as file2)
    For each pair, all variables are compared. Cases with differences are flagged by test variables and
    are saved in file 'records with differences.sav' (in same folder as file2).
    """

    tmp = os.path.split(file2)[0]
    dupFileSpecs = tmp + '/duplicate or missing keys.sav'
    deltaFileSpecs = tmp + '/records with differences.sav'

    numVarNames = ' '.join(varNumSet11)
    strVarNames = ' '.join(varStrSet11)
    keyNames = ' '.join(key)
    vnames = keyNames + ' ' + numVarNames + ' ' + strVarNames
    vnamesNoKey = numVarNames + ' ' + strVarNames

    nbNums = len(varNumSet11)
    nbStr = len(varStrSet11)
    nbTests = nbNums + nbStr    #number of variables to be tested
    sortCmd = 'SORT CASES BY %s.' % keyNames

    # Find cases with unmatched or duplicate keys

    cmd="""
    DATASET CLOSE ALL.

    GET FILE="%(file1)s" /KEEP=%(vnames)s.
    COMPUTE fileID = 1.
    %(sortCmd)s
    DATASET NAME file1.

    GET FILE="%(file2)s" /KEEP=%(vnames)s.
    COMPUTE fileID = 2.
    %(sortCmd)s

    ADD FILES FILE=file1
            /FILE=*
            /BY=%(keyNames)s fileID.
    MISSING VALUES %(vnames)s ().
    AGGREGATE
      /PRESORTED
      /BREAK = %(keyNames)s
      /nbLines = N
      /minFile = MIN(fileID)
      /maxFile = MAX(fileID).
    DATASET NAME new.
    DATASET CLOSE file1.
    COMPUTE filter$ = NOT ( nbLines = 2 AND minFile <> maxFile ) .
    FILTER BY filter$.
    SAVE OUTFILE='%(dupFileSpecs)s' /UNSELECTED=DELETE
        /DROP = minFile maxFile filter$. """ % locals()
    spss.Submit(cmd)

    # Define myText
    if nbTests == 0:
        spss.Submit("GET FILE='%s'." % dupFileSpecs)
        nbCasesDup = spss.GetCaseCount()
        if nbCasesDup:
            myText = ['%u cases do not match. Refer to file' % nbCasesDup,
                      '"%s"' % dupFileSpecs,
                      'for the list of these cases.']
        else:
            myText = ['The content of compatible variables contained in the 2 files are IDENTICAL']
        return '\n'.join(myText)
    else:
        defineLabs = ['/test%u "%s " '% (idx + 1, lab) for (idx,lab) in enumerate(vnamesNoKey.split(' '))]
        defineLabs = '\n '.join(defineLabs)

        cmd="""
            FILTER OFF.
            SELECT IF NOT filter$.

            VECTOR test(%(nbTests)s) .
            DO IF fileID=2.
              DO REPEAT idx=1 TO %(nbNums)s /var = %(numVarNames)s .
                  COMPUTE test(idx) = ( var = LAG(var) ) OR ( SYSMIS(var) AND SYSMIS(LAG(var))) .
              END REPEAT PRINT.
              DO REPEAT idx=1 TO %(nbStr)s /var = %(strVarNames)s.
                  COMPUTE test(idx + %(nbNums)s) = ( var = LAG(var) ) .
              END REPEAT PRINT.
            END IF.
            AGG OUTFILE=* MODE=ADDVARIABLES OVERWRITE=YES
              /BREAK=%(keyNames)s
              /test1 TO test%(nbTests)u = MAX( test1 TO test%(nbTests)u ) .

            COMPUTE nbDelta = %(nbTests)u -  SUM( test1 TO test%(nbTests)u ) .
            SELECT IF nbDelta > 0.
            VALUE LABELS test1 TO test%(nbTests)u 0 "Difference" 1 "ok ".
            VARIABLE LABEL nbDelta "Number of differences within a case" %(defineLabs)s .
            SAVE OUTFILE="%(deltaFileSpecs)s" /DROP=nbLines filter$ minFile maxFile.
            GET FILE="%(deltaFileSpecs)s" .
            """ % locals()

        print cmd   #to see faulty code should there be an error
        spss.Submit(cmd)
        allVarsInDeltaFile = spssaux.GetVariableNamesList()
        nbCasesDelta = spss.GetCaseCount()

        # Prepare myText summarizing findings
        spss.Submit(r"GET FILE='%s'." % dupFileSpecs)
        nbCasesDup = spss.GetCaseCount()
        if nbCasesDup:
            myText = ['%u cases have duplicate or unmatched keys. Refer to file' % nbCasesDup,
                      '"%s"' % dupFileSpecs,
                      'for the list of these cases.']
        else:
            myText = ['Using key = %s, every case from either file has a unique match in the other file' % keyNames]

        if nbCasesDelta > 0:     #There are paired keys with differences

            # Find test variables for which there are no differences between the keyed pairs
            cmd = r"""
                GET FILE="%(deltaFileSpecs)s".
                COMPUTE nobreak=1.
                AGG OUT=*
                  /BREAK=nobreak
                  /test1 TO test%(nbTests)u =SUM(test1 TO test%(nbTests)u)
                  /nbCases = N.
                FLIP.
                SELECT IF var001 < %(nbCasesDelta)u AND case_lbl <> "nobreak".
                FLIP.
                """ % locals()
            print cmd
            spss.Submit(cmd)
            allTestsVars = set(['test%u' % (n +1)  for n in range(nbTests)])
            vnames = spssaux.GetVariableNamesList()[1:] #exclude first var (case_lbl)
            vnamesSet = set(vnames) - set(['nbCases'])
            dropSet = allTestsVars - vnamesSet

            # Find variables corresponding to dropSet for which there are no differences

            for (idx,name) in enumerate(vnamesNoKey.split()):
                if 'test%u' % (idx + 1) in dropSet:
                    dropSet.add(name)
            dropString = ' '.join(dropSet)

            # Remove variables without differences

            vstring = ' '.join(vnamesSet)
            if dropString != '':
                dropString = '/DROP=' + dropString

            cmd = r"""
                GET FILE="%(deltaFileSpecs)s" %(dropString)s.
                CACHE.
                FREQ VAR= %(vstring)s .
                SAVE OUTFILE="%(deltaFileSpecs)s" .
                """ % locals()
            spss.Submit(cmd)
            print '/n**** vars with differences in "%s" :\n' % deltaFileSpecs,vnames
            myText.extend(['\n%u of the matched cases have differences. Refer to file' % nbCasesDelta,
                      '"%s"' % deltaFileSpecs ,
                      'for a complete listing.\n'])
        else:
            plural = ['','s'][len(key)>1]
            myText.extend(['Cases matched using the %s key%s variable%s were found to be IDENTICAL\n' % (keySource,plural,plural)])
        return '\n'.join(myText)

def _getSortedStr(iterObject):
    """input: an iterable objects
    returns: a sorted string (returns 'None' if string is empty)
    """
    tmp = list(iterObject)
    if tmp == []:
        return 'None'
    tmp.sort()
    return ' '.join(tmp)

def compare2files(file1, file2, key=None):
    """Compares 2 sav files and finds any dictionary differences and
    value differences
    file1 = filespecs of first file eg file1 = r'c:/temp/file1.sav'
    file2 = filespecs of second file
    key   = a list of one or more variables that uniquely identify each case
            eg key = 'eerID eeeID'.split() or
            key = 'countryID cityID householdID personID'.split()
            if key == None, key is set equal to (up to) 50 'compatible' variables present
            in both files ('compatible' means they are both numeric or both strings of same length)
    """

    print ['File1: "%s"\nFile2: "%s"\n\n' % (file1 ,file2)]
    assert os.path.isfile(file1) == True ,'File "%s" does not exist, please check filespecs' % file1
    assert os.path.isfile(file2) == True ,'File "%s" does not exist, please check filespecs' % file2
    assert key==None or len(key) < 51, 'key is currently limited to 50 variables'

    nbCases1, varSet1, fullDict1 = _getFullDict(file1)
    nbCases2, varSet2, fullDict2 = _getFullDict(file2)
    filespecs = ['File1: "%s"\nFile2: "%s"\nNumber of cases: File1(%u), File2(%u)\
        \n' % (file1, file2, nbCases1, nbCases2)]

    varNumSet1 = set([var for var in fullDict1 if fullDict1[var][0] == 0])
    varNumSet2 = set([var for var in fullDict2 if fullDict2[var][0] == 0])

    varSet10 = varSet1 - varSet2                #vars only in file1
    varSet01 = varSet2 - varSet1                #vars only in file2
    varSet00 = varSet10.union(varSet01)         #vars missing in one file
    varSet11 = varSet1.intersection(varSet2)    #vars exist in both files

    # The following dictionary items are compared
    desc = 'varType varFormat varLabel valueLabels missingValues varLevel'

    if not varSet11:        # file1 and file2 do not share any variables. Nothing left to check
        textVarsComp = ['The 2 files do not share any variable (nothing to check)\n']
        filespecs.extend(textVarsComp)
        procTitle = 'Compare 2 files (dictionary)'
        textTitle = "Occurrence of Variables in File1 and File2"
        _printResults(procTitle, textTitle, filespecs, None, desc)
        return

    #Find 'compatible' variables (numeric or strings of same length) in both files
    #These vars will be compared one by one for each case

    varStrSet11 = set([var for var in varSet11 if fullDict1[var][0] > 0 and \
                        fullDict1[var][0] == fullDict2[var][0]])
    varNumSet11 = varNumSet1.intersection(varNumSet2)     #vars numeric in both files
    keyText = 'Given Key vars: '
    msg = []
    keySource = 'Given'

    # Make tests on key

    keyVarsCompatible = True
    keyVarsPresent = True
    if key:
        for var in key:
            if (var in varSet1) and (var in varSet2):
                if fullDict1[var][0] != fullDict2[var][0]:
                    keyVarsCompatible = False
                    msg = ['\n\n*** key variable %s exists in both files but formats are NOT compatible.\
                           \nRefer to next table for complete list of variables with different format. \
                           \nCases cannot be matched using given keys ***\n' % var]
                    break
            else:       #Key vars are not present in both files
                keyVarsPresent = False
                msg = ['\n\n*** At least one key variable (%s) is NOT present in both files ***\n' % var]
                break
    else:   # Need to define assumed keys
        key = list(varNumSet11)[:50]
        nbKeys = len(key)
        if nbKeys < 50:
            key.extend(list(varStrSet11)[:50 - nbKeys])
        keyText = 'ASSUMED Key vars (none were supplied): '
        keySource = 'Assumed'

    varStrSet11 = varStrSet11 - set(key)
    varNumSet11 = varNumSet11 - set(key)

    textVarsComp = [keyText, _getSortedStr(key),
                    '\nVars only in file1: ', _getSortedStr(varSet10),
                    '\nVars only in file2: ', _getSortedStr(varSet01),
                    '\nVars in both files: ', _getSortedStr(varSet11),
                    '\nNum Vars in both files (excluding key vars): ' ,_getSortedStr(varNumSet11),
                    '\n']
    textVarsComp.extend(msg)

    #Check dict for vars in both files:

    myCells = []
    for id in varSet11:
        idstr = id
        for (d, p1,p2) in zip(desc.split(), fullDict1[id][0:5],fullDict2[id][0:5]):
            if p1 != p2:
                tmplist = [idstr, d, str(p1),str(p2)]
                myCells.extend(tmplist)
                idstr = ' '
    if not myCells:
        textVarsComp.extend(['\nDictionary information of vars in both files are IDENTICAL'])
    filespecs.extend(textVarsComp)
    procTitle = 'Compare 2 files (dictionary)'
    textTitle = "Occurrence of Variables in File1 and File2"
    _printResults(procTitle, textTitle, filespecs, myCells, desc)

    procTitle = 'Compare 2 files (case by case)'
    textTitle = "Summary of Totals and case per case comparisons"
    if not (keyVarsPresent and keyVarsCompatible):
        return  # Key error, exit program

    if nbCases1 == 0 or nbCases2 == 0 :
        myText = 'Since at least one of the files is empty, cases cannot be compared.'
        _printResults(procTitle, textTitle, myText, None, None)
        return  #Can't compare totals nor individual cases, exit program

    varNames = ' '.join(varNumSet1.intersection(varNumSet2))
    msgTotals = 'The 2 files do not share any numeric variables, cannot compare totals.'
    if varNames:
        #Compare Totals of Numeric variables present in both files
        msgTotals = _compareTotals(file1, file2, varNames, nbCases1, nbCases2)

    # Compare every value of all variables present in both files
    # This is done for each pair of records having a given key
    # Cases with differences are saved in file 'records with differences.sav'

    # If a key is present in only one file or is duplicated in a file, these cases
    # are saved to 'duplicate or missing keys.sav'

    #The above 2 files will be saved in the folder of file2

    msgCase = '\n'.join(['All (compatible) variables common to both files are included in the key.',
               'Hence, cases matched using the key must be equal (no need to check it).'])
    #if varNumSet11 or varStrSet11 or key:
        # if there are non-key variables present in both files then
        # match cases using key then compare values of non-key variables case per case
    msgCase = _compareCases(file1, file2, key, keySource, varNumSet11, varStrSet11)

    myText = msgTotals + '\n\n' + msgCase
    _printResults(procTitle, textTitle, myText, None, None)
    print '********* FINISHED *********'


def __test(file1, file2, key=None):
    # If a given test file does not exist, create it
    # (By not overwriting test file, script allows you to add additionnal differences
    # between the two files and test if the program detects them)

    if not os.path.isfile('C:/temp/eee data2.sav') or \
            not os.path.isfile('C:/temp/eee data2 noCases.sav'):
        spss.Submit(r"""
            SET PRINTBACK=YES /MPRINT=YES.
            GET FILE='c:/Program files/SPSS/employee data.sav'.
            IF prevexp=476 id=2000.
            MISSING VALUES salary().
            VARIABLE LABEL gender .
            STRING jobcat1(A1).
            COMPUTE jobcat1=STRING(jobcat,F1).
            RENAME VAR (salbegin jobcat1 jobcat = begsal jobcat jobcat1).
            DO IF id=10.
              COMPUTE salary = salary + 5000.
            ELSE IF id=11.
              COMPUTE salary = salary - 5000.
              COMPUTE educ = educ -1.
            END IF.
            COMPUTE var1 = (salary < 40000).
            SORT CASES BY gender.
            CACHE.
            SAVE OUTFILE='C:/temp/eee data2.sav' /DROP=jobcat1.
            SELECT IF id=3.
            ADD FILES FILE=*
                /FILE='C:/temp/eee data2.sav'.
            SAVE OUTFILE='C:/temp/eee data2.sav' /DROP=jobcat1.
            SELECT IF 1=0.
            SAVE OUTFILE='C:/temp/eee data2 noCases.sav' /DROP=jobcat1.
            """)

    if not os.path.isfile('c:/temp/copy of employee data.sav'):
        spss.Submit(r"""
            SET PRINTBACK=YES /MPRINT=YES.
            GET FILE='c:/Program files/SPSS/employee data.sav'.
            SAVE OUTFILE='c:/temp/copy of employee data.sav'. """ )

    if not os.path.isfile('c:/temp/copy of employee data.sav'):
        spss.Submit(r"""
            SET PRINTBACK=YES /MPRINT=YES.
            GET FILE='c:/Program files/SPSS/employee data.sav'.
            SAVE OUTFILE='c:/temp/copy of employee data.sav'. """ )

    compare2files(file1, file2, key)

if __name__ == '__main__':
    """ Create a test file (if required) to be able to compare """

    #List of test files:
    file1 = r'C:/Program files/SPSS/employee data.sav'  #This file is compared to the others
    file2 = r'C:/temp/eee data2.sav'                    #file1 with some modifications
    file3 = r'C:/Program files/SPSS/cars.sav'           #not related to file1
    file4 = r'C:/temp/eee data2 noCases.sav'            #file2 without any cases
    file5 = r'C:/temp/copy of employee data.sav'        #file5 = file1

    key = ['id']       #key needs to be None or a list of strings


    __test(file1, file2, None)          #assume there are no keys
    __test(file1, file2, key)
    __test(file1, file2, ['jobcat'])    #use wrong key
    __test(file1, file3, None)          #use wrong file
    __test(file1, file4, key)
    __test(file1, file5, key)
