#!/usr/bin/env python
"""
To standardize variable formats of various files.
For a given string variable, the new size is targetWidth (the maximum
size of this variable across all files matching the given pattern).
Numeric variables with the same name are converted to string using
STRING tmpVar(A targetWidth).
COMPUTE tmpVar = NUMBER(varname, Fw.d).
where w = min(targetWidth,40) and d = min(targetWidth-1,16)

Example of use:
# You should first save a copy of compare2files.py to your
# "C:\Python24\Lib\site-packages\" folder

BEGIN PROGRAM.
import standardizeVars as sv
sv.standardizeVars(drivepath = "c:/temp/",
               searchSubFolder = False,
               pattern = r"file\d+.sav",
               addFlag = True,
               runSyntax = False,
               newFileName = r'c:/temp/combined data.sav',
               newSyntaxName = r'c:/temp/combined data.sps')
END PROGRAM.

pattern is a RegEx, for examples:
  file\d+\.sav      all sav files whose name is 'file' followed by digits
  \bfile\d+\.sav    all sav files whose name starts with 'file' followed by digits
  .*\.sav           all sav files
  ^f1\.sav|^f2\.sav to search for 'f1.sav' and 'f2.sav'
"""

__author__  =  'Raynald Levesque raynald@spsstools.net'
__version__ =  '1.1'

import spss, spssaux, random, os, re, os.path

#History
# v 1.1 done 2007/04/28

# potential future improvements:
# suggestions are welcomed ...
# >>>>>> Test that any initial data set is protected

def createSampleFiles(nbFiles=4):
    """Creates sample files to illustrate use of standardizeVars()
    nbFiles is the number of files to create (minimum # is 4 (the default))
    Only 4 different files are created. When nbFiles > 4, the 4th file is repeated
    for the 5th and subsequent files.
    The main purpose of using nbFiles > 4 is to test that the ADD commands works
    correctly when there are more than 50 files.
    """

    # create C:/temp/ and C:/temp/test/ folders if they do not exist
    try:
        os.makedirs('C:/temp/test/')
    except:
        pass    # folder exists, nothing to do

    print '** making %(nbFiles)s dummy files.sav\n' % locals()
    # Note: syntax within next spss.Submit() command *must* start in first column
    # because of the BEGIN DATA and END DATA commands
    spss.Submit(r"""
SET PRINTBACK=YES /MPRINT=YES.
CD "c:/temp/".
DATA LIST LIST /var1(F2) var2(A8) var3(A3) var5(F8) var7(A12) var8(A7) date1(EDATE10).
BEGIN DATA
1 abc 1 5 aaaaaaa xyz 31-01-2007
END DATA.
SAVE OUTFILE="file1.sav".

DATA LIST LIST /var1(F2) var2(A6) var3(A3) var5(A8) var6(F4) var7(A12) var8(A7) date1(ADATE).
BEGIN DATA
1 abc 1 5 777 1222 aaaaa 02-28-2007
END DATA.
FORMATS date1(F8).
SAVE OUTFILE="file2.sav".

DATA LIST LIST /var1(F2) var3(A3) var5(A11) var6(A4) var7(A10) var8(A7).
BEGIN DATA
1 1 5 777 aaaaaaa oui
END DATA.
SAVE OUTFILE="fileABC.sav".

DATA LIST LIST /var1(F2) var2(A8) var3(A3) var5(A11) var6(A4) var7(A10) var8(A7).
BEGIN DATA
1 mno1 5 777 bbbb oui xyz
END DATA.
SAVE OUTFILE="test/file3.sav". """ )

    if nbFiles > 4:
        for idx in range(4,nbFiles):
            spss.Submit(r"""
            SAVE OUTFILE='test/file%s.sav'. """ % (idx))

def _getBasicInfo(drivepath="c:/temp/", searchSubFolder=False, pattern=r"*.sav"):
    """This function gets basic information about vars, format and location:
    It returns AllF which is a dictionnary of the following format:
        {'var1':{filename:format,...},
        {'var2':{filename:format,...},...}
    It also returns listSavFiles which is the list filespecs of all sav files
    """

    # protect active dataset
    dsn = spssaux.GetActiveDatasetName()
    if not dsn:
        spss.Submit("DATASET NAME S%s." % str(random.uniform(0,1)) )

    allF = {}
    sumList = {}
    pat = re.compile(pattern, re.IGNORECASE)
    msg='folder '
    if searchSubFolder: msg = 'directory tree '
    print "searching " + msg + drivepath + " and filenames like " + pattern

    f = os.walk(drivepath)  #create generator function
    for (dirname, subdirs, files) in f:
        if dirname == drivepath or searchSubFolder == True:
            for sav in files:
                if pat.match(sav):
                    try:
                        spss.Submit('get file="%s/%s".' % (dirname,sav))
                        varDict = spssaux.VariableDict()
                        if dirname == drivepath:
                            fullname = dirname + sav
                        else:
                            fullname = dirname + '/' + sav
                        print 'Found:',fullname
                        for id in varDict.variables:
                            try:
                                allF[id][fullname]=(varDict.VariableFormat(id),varDict.VariableType(id))
                            except:
                                allF[id]={fullname:(varDict.VariableFormat(id),varDict.VariableType(id)) }
                        del varDict
                    except:
                        print 'file="%s/%s" is not a sav file.' % (dirname,sav)
                        pass    #This is not an SPSS file, ignore it
                    try:
                        listSavFiles.append(fullname)
                    except:
                        listSavFiles = [fullname]
    try:
        return allF, listSavFiles
    except:             # No files were found
        return {},{}

def _eliminateOK(varList):
    """Finds across-file status (Mixed, Strings or OK) of every variable
    Removes variables from varList that are already consistent across files
    Returns varList with for each vars, it's status, actual and target formats
    """

    print '\n*****Find across-file status (Mixed, Strings or OK) of every variable:'
    delList = ''
    reformatList = ''
    mixedList = ''
    for var in varList:
        print
        for file in varList[var]:
            try:
                minVal = min(minVal,varList[var][file][1])
                maxVal = max(maxVal,varList[var][file][1])
            except:
                minVal = varList[var][file][1]
                maxVal = varList[var][file][1]
            print var, file, varList[var][file], minVal,maxVal

        if minVal == maxVal:
            print '%(var)s is already OK, place it on delete list' % locals()
            delList = delList + ' ' + str(var)
            varList[var]['status'] = 'OK'
        elif minVal > 0:
            print 'need to standardize width of these strings vars'
            reformatList = reformatList + ' ' + str(var)
            varList[var]['status'] = 'Strings'
        else:
            print 'some are strings, some are numerics'
            mixedList = mixedList + ' ' + str(var)
            varList[var]['status'] = 'Mixed'
        print 'status = ', varList[var]['status']
        varList[var]['maxLength'] = (maxVal)
        del minVal,maxVal

    print '\nRemove vars that are consistent across all files:'
    for var in delList.split():
        print '       ',var,
        del varList[var]
    return varList


def _getFileList(varList):
    """Takes varList which is a dictionnary of the following format:
    {'var1':{filename:format,...},
    {'var2':{filename:format,...},...}
    and creates a fileList which is a new dictionnary with the following format:
    {'filename1':{var:format,...},
    {'filename2':{var:format,...},...}
    """
    fileList = {}
    for var in varList:
        for file in varList[var]:
            if file != 'maxLength' and file != 'status':
                try:
                    fileList[file][var] = (varList[var]['status'],
                                           varList[var]['maxLength'],
                                           varList[var][file][1])
                except:
                    fileList[file] = {var:(varList[var]['status'],
                                           varList[var]['maxLength'],
                                            varList[var][file][1]) }
    return fileList


def _ChangeStringWidth(varname, currentWidth, targetWidth, cmd):
    """Adds syntax to cmd, to
    1)Change the width of an SPSS string variable, or
    2)Convert a numeric variable to a string variable
    varname is the name of the variable
    currentWidth of varname (0 means it is a numeric var)
    targetWidth of varname (when current == target, there is nothing to do)
    cmd is current syntax
    Function returns the new cmd
    """
    tempName="T" + str(random.random())
    if currentWidth > 0:    # This is a string var
        width = int(targetWidth) or 1
        cmd.extend([r"""
        STRING %(tempName)s (A%(targetWidth)s).
        COMPUTE %(tempName)s = %(varname)s.
        APPLY DICTIONARY FROM=*
               /SOURCE VARIABLES=%(varname)s
               /TARGET VARIABLES=%(tempName)s
               /VARINFO ATTRIBUTES LEVEL MISSING VALLABELS VARLABEL.
        MATCH FILES FILE=* /DROP= %(varname)s.
        RENAME VARIABLE (%(tempName)s=%(varname)s).
        """ %locals()])
    else:                   # This is a numeric var
        target2 = min(targetWidth - 1,16)
        cmd.extend([r"""
        STRING %(tempName)s (A%(targetWidth)s).
        COMPUTE %(tempName)s = STRING(%(varname)s,F%(targetWidth)s.%(target2)s).
        APPLY DICTIONARY FROM=*
               /SOURCE VARIABLES=%(varname)s
               /TARGET VARIABLES=%(tempName)s
               /VARINFO VARLABEL.
        MATCH FILES FILE=* /DROP= %(varname)s.
        RENAME VARIABLE (%(tempName)s=%(varname)s).
        """ %locals() ])
        pass
    return


def _stdStrings(drivepath, fileList, listSavFiles, protectFiles, addFlag, defineSourceLabels ):
    """Standardize length of strings variables
    Converts numeric variables to strings variables with targetWidth
    This is called by standardizeVars()
    """
    cmd = ['* Syntax written by stdStrings().']
    if protectFiles:
        try:
            pathtmp = 'C:/temp/T%s/' % str(random.uniform(0, 1))
            os.makedirs(pathtmp)
        except:
            pass    # folder exists, nothing to do

    smallSet = set(fileList)
    largeSet = set(listSavFiles)
    noDeltaList = list(largeSet - smallSet)
    sizeSmallSet = len(smallSet)

    # Process files containing variables that are not compatible
    for (idx,file) in enumerate(fileList):
        idx2 = idx + 1
        firstVar = True
        defineSourceLabels.extend([' %u "%s" ' % (idx2, file)])
        for var in fileList[file]:
            targetWidth = fileList[file][var][1]
            currentWidth = fileList[file][var][2]
            if currentWidth < targetWidth:
                if firstVar:
                    cmd.extend(['\n* Doing next file *.',
                                'GET FILE="%s". ' % (file)])
                    firstVar = False
                _ChangeStringWidth(var, currentWidth, targetWidth, cmd)
        if firstVar == False:
            _saveFile(idx2, file, drivepath, pathtmp, protectFiles, cmd, listSavFiles)

    # Need to process files already compatible to add the source variable
    if addFlag:
        for (idx,file) in enumerate(noDeltaList):
            idx2 = idx + 1 + sizeSmallSet
            defineSourceLabels.extend([' %u "%s" ' % (idx2, file)])
            cmd.extend(['* Doing next file *.',
                        'GET FILE="%s". ' % (file)])
            _saveFile(idx2, file, drivepath, pathtmp, protectFiles, cmd, listSavFiles)
    return cmd

def _saveFile(idx, file, drivepath, pathtmp, protectFiles, cmd, listSavFiles):
    """ Write syntax to save file and create a source variable"""

    filespecs = file
    if protectFiles:
        filespecs = file.replace(drivepath, pathtmp)
    cmd.extend(['COMPUTE source = %u .' % idx,
           "SAVE OUTFILE='%s'. " % (filespecs)])

    # Update filespecs in listSavFiles
    idx = listSavFiles.index(file)
    listSavFiles[idx] = filespecs
    pathdir = os.path.dirname(filespecs)
    try:
        os.makedirs(pathdir)
    except:
        pass    # folder exists, nothing to do
    return


def _addFiles(listSavFiles, newFileName, cmd, defineSourceLabels):
    """writes syntax to add many sav files
    listSavFiles: a list of complete filepecs (i.e. paths and names) of sav files to be added
    newFileName:  the filespec of the new file eg 'c:\data\combined data.sav'
    cmd is the current syntax (command string)
    """
    first = True
    if len(listSavFiles) > 1:
        while listSavFiles:
            if first:
                cmd.extend(["\n\n* Add Files.\nADD FILES " + ' '.join(["\n /FILE='%s'" % (fn) for fn in listSavFiles[:50]]) + ".\n"])
                listSavFiles = listSavFiles[50:]
                first = False
            else:
                cmd.extend(["EXECUTE.\nADD FILES /FILE=*" + ' '.join(["\n /FILE='%s'" % (fn) for fn in listSavFiles[:49]]) + ".\n"])
                listSavFiles = listSavFiles[49:]
        cmd.extend(defineSourceLabels)
        cmd.extend(['\nCACHE.\nSAVE OUTFILE="%s".' % (newFileName)])
    else:
        cmd.extend(['*There are no files to ADD.\n'])
    return


def standardizeVars(drivepath="c:/temp/", searchSubFolder=False, pattern=r"file*.sav", protectFiles=True,
                    addFlag=False, newFileName=None, runSyntax=False, newSyntaxName=None):
    """This function is used to write (or run) a syntax that makes variables consistent across files
    so that ADD FILES can be used. If a variable occurs in more than 1 file, and has different format
    (eg string vars of different length or var is numeric in some files and string in other files,
    then the function writes syntax that makes these variables consistent.

    Parameters:
    drivepath       folder where search starts
    searchSubFolder True or False (sub-folders of drivepath are only searched
                    if True
    pattern         a RegEx of file names to search and process
    protectFiles    leave original sav files unchanged
    addFlag         If True, syntax to ADD all files is created. The resulting combined file
                    includes a source variable containing the filespecs of the source data file.
    newFileName     the result of the ADD command is saved with this filespec
    runSyntax       if True, final syntax is submitted to SPSS and executed
    newSyntaxName   filespec of the syntax created by this function
    """

    allFiles, listSavFiles = _getBasicInfo(drivepath=drivepath, searchSubFolder=searchSubFolder, pattern=pattern)
    if len(allFiles) > 0:
        defineSourceLabels = ['VALUE LABELS source']
        varList = _eliminateOK(allFiles)
        if len(varList)>0:
            print '\n**** REMAINING var list: ****'
            for var in varList:
                print '\n',var, 'status :', varList[var]['status']
                print '    maxLength:', varList[var]['maxLength']
                for file in varList[var]:
                    if file != 'status' and file != 'maxLength':
                        print '   ', file, varList[var][file][0],varList[var][file][1]

            fileList = _getFileList(varList)
            print '\nfileList =\n'
            for file in fileList:
                print file
                for var in fileList[file]:
                    print '     ', var, fileList[file][var][0], '\ttarget=',fileList[file][var][1], '\tcurrent=',fileList[file][var][2]
            cmd = _stdStrings(drivepath, fileList, listSavFiles, protectFiles, addFlag, defineSourceLabels)
            print 'Resulting command is:\n', '\n'.join(cmd)

            if addFlag:
                # zzz cmd = _addFiles(listSavFiles, newFileName, cmd, defineSourceLabels)
                _addFiles(listSavFiles, newFileName, cmd, defineSourceLabels)

            cmd = '\n'.join(cmd)
            if runSyntax:
                spss.Submit(cmd)
            else:
                print 'Final syntax file is :\n', cmd
        else:
            cmd = '\n* All %s files are already compatible.' % len(listSavFiles)
            print cmd
    else:
        cmd = '\n* No files were found.'
        print cmd

    if newSyntaxName:
        f = open(newSyntaxName,'w')
        f.write(cmd)
        f.close
        print '\n***File\n',newSyntaxName,'\ncontains the syntax generated by the program'

if __name__ == '__main__':
    # Test that the function works
    #createSampleFiles(nbFiles=4)
    standardizeVars(drivepath = "c:/temp/",
                searchSubFolder = True,
                pattern = r".*file\d+\.sav",
                addFlag = True,
                runSyntax = False,
                newFileName = r'c:\temp\combined data.sav',
                newSyntaxName = r'c:\temp\combined data.sps')
